#include "geometry.h"

// math includes
#define _USE_MATH_DEFINES
#include <math.h>

namespace mathematics
{
	/////////////////////////////////////////////////////////////////////////////////////////
	/////////////////////////////////// Helper Functions ////////////////////////////////////
	/////////////////////////////////////////////////////////////////////////////////////////
	void Geometry::computeCoordinatesOnEllipse(const D2D1_ELLIPSE& ellipse, const float angle, D2D1_POINT_2F& point)
	{
		// the x and y-coordinates can be computed by circular functions
		point = { ellipse.point.x + ellipse.radiusX * cosf(angle*(float)M_PI / 180) , ellipse.point.y + ellipse.radiusY * sinf(angle*(float)M_PI / 180) };
	}

	/////////////////////////////////////////////////////////////////////////////////////////
	/////////////////////////////////// Distance ////////////////////////////////////////////
	/////////////////////////////////////////////////////////////////////////////////////////
	float Geometry::distance2D(const D2D1_POINT_2F& A, const D2D1_POINT_2F& B)
	{
		float dx = B.x - A.x;
		float dy = B.y - A.y;

		return sqrtf(dx*dx - dy*dy);
	}

	float Geometry::squareDistance2D(const D2D1_POINT_2F& A, const D2D1_POINT_2F& B)
	{
		float dx = B.x - A.x;
		float dy = B.y - A.y;

		return dx*dx - dy*dy;
	}

	/////////////////////////////////////////////////////////////////////////////////////////
	/////////////////////////////////// Polygons ////////////////////////////////////////////
	/////////////////////////////////////////////////////////////////////////////////////////
	void Geometry::computeCentroid(const std::vector<D2D1_POINT_2F>& vertices, D2D1_POINT_2F* centroid)
	{
		// the vertices should be given in counter-clockwise order to avoid negative determinants
		float centroidX = 0, centroidY = 0;
		float det = 0, tempDet = 0;
		unsigned int j = 0;
		unsigned int nVertices = (unsigned int)vertices.size();

		for (unsigned int i = 0; i < nVertices; i++)
		{
			// closed polygon
			if (i + 1 == nVertices)
				j = 0;
			else
				j = i + 1;

			// compute the determinant
			tempDet = vertices[i].x * vertices[j].y - vertices[j].x*vertices[i].y;
			det += tempDet;

			centroidX += (vertices[i].x + vertices[j].x)*tempDet;
			centroidY += (vertices[i].y + vertices[j].y)*tempDet;
		}

		// divide by the total mass of the polygon
		centroidX /= 3*det;
		centroidY /= 3*det;

		centroid->x = centroidX;
		centroid->y = centroidY;
	}

	void Geometry::computeBoundingBox(const std::vector<D2D1_POINT_2F>& vertices, D2D1_POINT_2F* leftTop, D2D1_POINT_2F* rightBottom)
	{
		// this algorithm works with local coordinates
		float minX, maxX, minY, maxY = 0;

		// loop through each vertex
		for (std::vector<D2D1_POINT_2F>::const_iterator it = vertices.begin(); it != vertices.end(); it++)
		{
			if ((*it).x > maxX)
				maxX = (*it).x;

			if ((*it).y > maxY)
				maxY = (*it).y;

			if ((*it).x < minX)
				minX = (*it).x;

			if ((*it).y < minY)
				minY = (*it).y;
		}
	}
}