// INCLUDES /////////////////////////////////////////////////////////////////////////////

// c++ includes: streams
#include <sstream>

// class header
#include "playState.h"

// bell0bytes core
#include "app.h"
#include "depesche.h"
#include "coreComponent.h"
#include "window.h"

// bell0bytes file system
#include "fileSystemComponent.h"
#include "folders.h"

// bell0bytes input
#include "gameCommands.h"
#include "inputComponent.h"
#include "inputHandler.h"

// bell0bytes graphics
#include "graphicsComponent.h"
#include "graphicsComponent2D.h"
#include "d2d.h"
#include "graphicsComponentWrite.h"
#include "sprites.h"

// bell0bytes audio
#include "audioComponent.h"

// bell0bytes mathematics
#include "geometry.h"

// bell0bytes physics
#include "kinematics.h"

// CLASS METHODS ////////////////////////////////////////////////////////////////////////
namespace game
{
	/////////////////////////////////////////////////////////////////////////////////////////
	///////////////////////////// Constructor and Destructor ////////////////////////////////
	/////////////////////////////////////////////////////////////////////////////////////////
	PlayState::PlayState(core::DirectXApp& app, const std::wstring& name) : GameState(app, name)
	{ }
	PlayState::~PlayState()
	{ }
	PlayState& PlayState::createInstance(core::DirectXApp& app, const std::wstring& stateName)
	{
		static PlayState instance(app, stateName);
		return instance;
	}

	/////////////////////////////////////////////////////////////////////////////////////////
	/////////////////////////////// Initialization //////////////////////////////////////////
	/////////////////////////////////////////////////////////////////////////////////////////
	util::Expected<void> PlayState::initialize()
	{
		// catch errors
		util::Expected<void> result;
		
		// allow only keyboard input
		dxApp.getInputComponent().getInputHandler().activeKeyboard = true;
		dxApp.getInputComponent().getInputHandler().activeMouse = true;

		// set initial positions of the circle centers
		for (unsigned int i = 0; i < 50; i++)
		{
			centers.push_back(D2D1::Point2F(rand() % 1200, rand() % 500));
			centerVecs.push_back(mathematics::Vector2F(centers[i].x, centers[i].y));
		}
		float mpp = physics::Kinematics::metersPerPixel;

		// set initial velocities
		for (unsigned int i = 0; i < 50; i++)
			velocities.push_back(mathematics::Vector2F((rand()%250-125)*mpp, (rand() % 250 - 125)*mpp));

		// set initial accelerations
		for (unsigned int i = 0; i < 50; i++)
			accelerations.push_back(mathematics::Vector2F((rand()%20-10)*mpp,(rand()%20-10)*mpp));


		// notify the main application class that the game is running
		isPaused = false;

		firstCreation = false;
		
		// return success
		return { };
	}

	/////////////////////////////////////////////////////////////////////////////////////////
	///////////////////////////////  Pause and Resume ///////////////////////////////////////
	/////////////////////////////////////////////////////////////////////////////////////////
	util::Expected<void> PlayState::pause()
	{
		isPaused = true; 
		return { };
	}
	
	util::Expected<void> PlayState::resume()
	{
		isPaused = false;
		return { };
	}
	

	/////////////////////////////////////////////////////////////////////////////////////////
	/////////////////////////////// User Input //////////////////////////////////////////////
	/////////////////////////////////////////////////////////////////////////////////////////
	util::Expected<void> PlayState::handleInput(std::unordered_map<input::GameCommands, input::GameCommand&>& activeKeyMap)
	{
		// act on user input
		for (auto x : activeKeyMap)
		{
			switch (x.first)
			{

			case input::GameCommands::MoveLeft:
				break;

			case input::GameCommands::MoveRight:
				break;

			case input::GameCommands::MoveDown:
				break;

			case input::GameCommands::Back:
				dxApp.quitGame();
				break;

			default:
				break;
			}
		}
		return { };	
	}

	util::Expected<void> PlayState::onMessage(const core::Depesche& depesche)
	{
		if (depesche.type == core::DepescheTypes::ActiveKeyMap)
		{
			input::InputHandler* ih = (input::InputHandler*)depesche.sender;

			if (!isPaused)
				if (!ih->isListening())
					return handleInput(ih->activeKeyMap);
		}
		
		// return success
		return { };
	}

	/////////////////////////////////////////////////////////////////////////////////////////
	//////////////////////////////////// Update /////////////////////////////////////////////
	/////////////////////////////////////////////////////////////////////////////////////////
	util::Expected<void> PlayState::update(const double deltaTime)
	{
		if (isPaused)
			return { };

		// get meters per pixel
		const float mpp = physics::Kinematics::metersPerPixel;

		// update position based on symplectic integration of the equations of motion
		for (unsigned int i = 0; i < centers.size(); i++)
		{
			// symplectic integrator: semi-implicit Euler
			physics::Kinematics::semiImplicitEuler(centerVecs[i], velocities[i], accelerations[i], deltaTime);
			centers[i].x = centerVecs[i].x;
			centers[i].y = centerVecs[i].y;
			
			// wrap screen
			unsigned int width = dxApp.getGraphicsComponent().getCurrentWidth();
			unsigned int height = dxApp.getGraphicsComponent().getCurrentHeight();

			if (centers[i].x > width)
			{
				centers[i].x = width;
				accelerations[i].x = -accelerations[i].x;
				velocities[i].x = -velocities[i].x;
			}

			if (centers[i].x < 0)
			{
				centers[i].x = 0.0f;
				accelerations[i].x = -accelerations[i].x;
				velocities[i].x = -velocities[i].x;
			}

			if (centers[i].y < 0)
			{
				centers[i].y = 0.0f;
				accelerations[i].y = -accelerations[i].y;
				velocities[i].y = -velocities[i].y;
			}

			if (centers[i].y > height)
			{
				centers[i].y = height;
				accelerations[i].y = -accelerations[i].y;
				velocities[i].y = -velocities[i].y;
			}
		}

		// return success
		return { };
	}

	/////////////////////////////////////////////////////////////////////////////////////////
	/////////////////////////////////// Render //////////////////////////////////////////////
	/////////////////////////////////////////////////////////////////////////////////////////
	util::Expected<void> PlayState::render(const double farSeer)
	{
		// update positions if necessary
		for (unsigned int i = 0; i < centers.size(); i++)
		{
			centerVecs[i] += velocities[i] * farSeer;
			centers[i].x = centerVecs[i].x;
			centers[i].y = centerVecs[i].y;
		}

		// render circles
		for (unsigned int i = 0; i < centers.size(); i++)
			dxApp.getGraphicsComponent().get2DComponent().drawEllipse(centers[i].x, centers[i].y, 10.0f, 2.5f);

		// print FPS information
		dxApp.getGraphicsComponent().getWriteComponent().printFPS();

		// return success
		return { };
	}

	/////////////////////////////////////////////////////////////////////////////////////////
	/////////////////////////////// Shutdown ////////////////////////////////////////////////
	/////////////////////////////////////////////////////////////////////////////////////////
	util::Expected<void> PlayState::shutdown()
	{
		isPaused = true;

		// return success
		return { };
	}
}